#ifndef FZSSH_VISIBILITY_CRYPT_HEADER
#define FZSSH_VISIBILITY_CRYPT_HEADER

#include <libfilezilla/visibility_helper.hpp>

#ifdef BUILDING_FZSSH_CRYPT
  #define FZSSH_CRYPT_PUBLIC_SYMBOL FZ_EXPORT_PUBLIC
  #define FZSSH_CRYPT_PRIVATE_SYMBOL FZ_EXPORT_PRIVATE
#else
  #define FZSSH_CRYPT_PUBLIC_SYMBOL FZ_IMPORT_SHARED
  #define FZSSH_CRYPT_PRIVATE_SYMBOL
#endif

#endif
